import random
import numpy as np
from functools import partial
from multiprocessing import Pool, cpu_count

from deap import base
from deap import creator
from deap import tools

from .LocalOptimizer import optimizer
from .Tournament import Tournament
from .InitialSolution import initial_solution
from .Mutation import mutation
from .Recombination import recombination
from .Score import score
from deap.tools.indicator import hv


def helper(func, ind, rnd):
    if rnd < 0.1:
        func(ind)
    return ind


class TSM(object):
    def __init__(self, data, data2, overview, solutions, tour, MU, m_pr, c_pr, prob, weights1, ref):
        """
        Hybrid GA (memetic) for the FJSSP
        :param data:
        :param data2:
        :param overview:
        :param solutions:
        :param MU:
        :param m_pr:
        :param c_pr:
        :param weights1:
        :param prob:
        :param kwargs:
        """
        self.MU = MU
        self.tour = tour
        self.solutions = solutions
        self.ref = ref
        self.weights1 = weights1
        self.m_pr = m_pr
        self.c_pr = c_pr
        self.prob = prob
        self.data = data
        self.data2 = data2
        self.overview = overview
        self.max_mach = len(overview[0][1])
        self.job_n = len(overview)
        self.data3 = dict()
        for key in self.data2.keys():
            self.data3[key] = key[0]
        self.x = None
        self.y = None
        self.toolbox = None
        self.ex_solutions = 0

    def main(self):
        creator.create("FitnessMin", base.Fitness, weights=self.weights1)  # minimizing bi-objective
        creator.create("Individual", list, fitness=creator.FitnessMin)

        self.toolbox = base.Toolbox()
        self.toolbox.register("init", initial_solution, self.data, self.data2, self.overview, self.job_n)
        self.toolbox.register("individual", tools.initRepeat, creator.Individual, self.toolbox.init, n=1)
        self.toolbox.register("population", tools.initRepeat, list, self.toolbox.individual)
        self.toolbox.register("mate", recombination, self.overview, self.job_n)
        self.toolbox.register("evaluate", score, self.data, self.overview, self.max_mach)
        self.toolbox.register("parents", Tournament)
        self.toolbox.register("mutate", mutation, self.overview, self.max_mach, self.data, self.data2, self.job_n)
        self.toolbox.register("opt", optimizer, self.data, self.data2, self.job_n, self.max_mach,
                              self.overview)
        self.toolbox.register("sort", tools.sortLogNondominated)
        self.toolbox.register("select", tools.selNSGA2)

        a, b, c = self.run()
        return a, b, c

    def run(self):
        no_prog = 0
        gens = 0
        print('generation ' + str(gens))

        "Initialize population and evaluate"
        with Pool(cpu_count()) as p:
            pop = p.map(self.toolbox.population, [self.MU])[0]

        self.ex_solutions = self.MU
        with Pool(cpu_count()) as p:
            pop = p.starmap(partial(helper, self.toolbox.opt), [(ind, random.random()) for ind in pop])

        fitnesses = [self.toolbox.evaluate(x) for x in pop]
        A = zip(pop, fitnesses)
        for ind, fit in A:
            ind.fitness.values = fit

        fronts = self.toolbox.sort(pop, len(pop))
        fits = [ind.fitness.values for ind in fronts[0]]
        archive = [fits]

        print(fits)
        HV = hv.hypervolume(fits, self.ref)
        HV_old = HV

        "Begin the evolution"

        while self.ex_solutions < self.solutions:

            gens += 1
            print('generation ' + str(gens))

            all_fit = [ind.fitness.values for ind in pop]
            similar = [[x for x in pop if x.fitness.values == y] for y in all_fit]
            ind_max_similar = np.argmax([len(x) for x in similar])
            for ind in similar[ind_max_similar]:
                self.ex_solutions += 1
                self.toolbox.mutate(ind)
                del ind.fitness.values
                ind.fitness.values = self.toolbox.evaluate(ind)

            offspring, ex_solutions = genetic_operators(pop, self.toolbox, self.MU, self.c_pr, self.m_pr, self.prob,
                                                        self.tour)

            pop = pop + offspring

            fronts = self.toolbox.sort(pop, len(pop))
            archive.append([ind.fitness.values for ind in fronts[0]])
            pop = self.toolbox.select(pop, self.MU)

            self.ex_solutions += ex_solutions
            print(self.ex_solutions)
            fronts = self.toolbox.sort(pop, len(pop))
            fits = [ind.fitness.values for ind in fronts[0]]
            HV = hv.hypervolume(fits, self.ref)

            if HV == HV_old:
                no_prog += 1
            else:
                no_prog = 0
                HV_old = HV
            if no_prog > 3:
                self.prob = 0.5
            else:
                self.prob = 0.1

        return archive, fits, pop


def genetic_operators(pop, toolbox, MU, c_pr, m_pr, prob, tour, parallel=True):
    """
    Helper function for parallelization
    :param pop:
    :param toolbox:
    :param MU:
    :param c_pr:
    :param m_pr:
    :param prob:
    :param parallel:
    :return:
    """
    offspring = toolbox.parents(pop, MU, tour)

    offspring = list(map(toolbox.clone, offspring))
    for child1, child2 in zip(offspring[::2], offspring[1::2]):
        if random.random() < c_pr:
            toolbox.mate(child1, child2)
            del child1.fitness.values
            del child2.fitness.values

    do_individual = partial(update_individual, m_pr=m_pr, toolbox=toolbox, prob=prob)

    if parallel:
        with Pool(cpu_count()) as p:
            out = p.map(do_individual, offspring)
    else:
        out = map(do_individual, offspring)

    children = [x[0] for x in out]
    ex_solutions = [x[1] for x in out]
    return children, sum(ex_solutions)


def update_individual(ind, m_pr, toolbox, prob):
    ex_solutions = 0
    if random.random() < m_pr:
        toolbox.mutate(ind)
        del ind.fitness.values
    if random.random() < prob:
        _, ex_solutions = toolbox.opt(ind)
        del ind.fitness.values
    if not ind.fitness.valid:
        ex_solutions += 1
        ind.fitness.values = toolbox.evaluate(ind)
    return ind, ex_solutions
